# Copyright 2006 Lars Wirzenius <liw@iki.fi>
#
# This program is free software; you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation; either version 2 of the License, or
# (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program; if not, write to the Free Software
# Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA


import md5
import os
import re
import sets
import sha
import stat
import sys
import time
import UserDict
import UserList


from simpletask import BackgroundTask, TaskQueue


"""User interface independent library stuff for Summain

Lars Wirzenius <liw@iki.fi>
"""


class FileFinder(BackgroundTask):

    """Find files in directory trees"""
    
    def __init__(self, names):
        self._remaining_names = names[:]
        self._result = []
    
    def get_result(self):
        """Return filenames + stat results found so far"""
        return self._result
        
    def estimate(self):
        if self.is_running():
            return 0, len(self._result)
        else:
            return len(self._result), len(self._result)

    def work(self):
        if self._remaining_names:
            name = self._remaining_names[0]
            del self._remaining_names[0]
            try:
                st = os.stat(name)
            except os.error:
                # Something bad happened, ignore the file.
                # FIXME: Should report this somehow to the user
                pass
            else:
                if stat.S_ISDIR(st.st_mode):
                    try:
                        basenames = os.listdir(name)
                    except os.error:
                        # Unreadable directory, or other problems.
                        # FIXME: Should report to user somehow.
                        pass
                    else:
                        for basename in basenames:
                            fullname = os.path.join(name, basename)
                            self._remaining_names.insert(0, fullname)
                elif stat.S_ISREG(st.st_mode):
                    self._result.append((name, st))
                else:
                    # Not a plain file, ignore it.
                    # FIXME: Should report this somehow to the user
                    pass

        return self._remaining_names != []


class ManifestFile(UserDict.UserDict):

    """A collection of data about one file in a Manifest
    
    This is a dictionary, indexed by the checksum name, and also
    has the ".filename" member to hold the filename, and the 
    ".st" member for the result of os.stat.
    
    """
    
    def __init__(self, filename, st):
        UserDict.UserDict.__init__(self)
        self.filenamex = filename
        self.st = st


class Manifest:

    """A virtual base class for manifest files"""
    
    # We're not a UserList, since we need an incompatible constructor.
    
    _type = None
    
    def __init__(self, base):
        self._list = []
        self._base = base
    
    def __getitem__(self, key):
        return self._list[key]
        
    def __setitem__(self, key, value):
        self._list[key] = value
        
    def __delitem__(self, key):
        del self._list[key]
        
    def __iter__(self):
        return iter(self._list)
        
    def __contains__(self, item):
        return item in self._list
    
    def append(self, item):
        self._list.append(item)
        
    def insert(self, key, item):
        self._list.insert(key, item)
        
    def remove(self, item):
        self._list.remove(item)
        
    def __len__(self):
        return len(self._list)
    
    def get_on_disk_filename(self, mf):
        return os.path.join(self._base, mf.filenamex)

    def get_in_manifest_filename(self, mf):
        if mf.filenamex is None:
            return "-"
        else:
            return mf.filenamex

    def write(self, output):
        """Write the manifest to an open output file"""
        raise Exception("Not implemented")

    def write_to_named_file(self, filename):
        """Write the manifest to a named file"""
        f = file(filename, "w")
        self.write(f)
        f.close()


class Md5Manifest(Manifest):

    """A manifest file with MD5 checksums ('md5sum.txt')"""

    _type = "md5"
    
    def write(self, output):
        for mf in self:
            assert "md5" in mf
            output.write("%s  %s\n" % 
                         (mf["md5"], self.get_in_manifest_filename(mf)))


class Sha1Manifest(Manifest):

    """A manifest file with SHA-1 checksums"""
    
    _type = "sha1"

    def write(self, output):
        for mf in self:
            assert "sha1" in mf
            output.write("%s  %s\n" % 
                         (mf["sha1"], self.get_in_manifest_filename(mf)))


class ManifestFactory:

    """Create a manifest object of the desired type"""

    _manifest_classes = {
        "md5": Md5Manifest,
        "sha1": Sha1Manifest,
    }
    
    def __init__(self, type_name):
        if type_name in self._manifest_classes:
            self._manifest_class = self._manifest_classes[type_name]
        else:
            raise Exception("Unknown manifest type '%s'" % type_name)

    def get_manifest(self, base):
        return self._manifest_class(base)


class ChecksumAlgorithmFactory:

    """Create an algorithm object for the desired checksum"""
    
    _algorithms = {
        "md5": md5,
        "sha1": sha,
    }

    def __init__(self, algorithm_name):
        if algorithm_name in self._algorithms:
            self._algorithm = self._algorithms[algorithm_name]
            self.name = algorithm_name
        else:
            raise Exception("Unknown algorithm '%s'" % algorithm_name)

    def new(self):
        return self._algorithm.new()


class FileChecksumComputer(BackgroundTask):

    """Compute the desired checksums of a file on disk"""
    
    _amount = 65536
    
    def __init__(self, manifest, filename, st, callback):
        self._manifest = manifest
        self._mf = ManifestFile(filename, st)
        self._callback = callback
        self._algos = []
        self._file = None
        self._bytes_read = 0
        
    def add_algorithm(self, name, algo):
        self._algos.append((name, algo))

    def setup(self):
        if self._mf.filenamex is None:
            self._file = sys.stdin
        else:
            filename = self._manifest.get_on_disk_filename(self._mf)
            self._file = file(filename, "r")
        
    def estimate(self):
        if self._mf.st is None:
            return self._bytes_read, self._bytes_read
        else:
            return self._mf.st.st_size, self._bytes_read

    def work(self):
        data = self._file.read(self._amount)
        if data:
            for name, algo in self._algos:
                algo.update(data)
            self._bytes_read += len(data)
            return True
        else:
            if self._file != sys.stdin:
                self._file.close()
            self._file = None
            return False

    def cleanup(self):
        for name, algo in self._algos:
            self._mf[name] = algo.hexdigest()
        self._callback(self._mf)


class FileChecksumComputerFactory:

    """Create a FileChecksumComputer object with the desired algorithms"""
    
    def __init__(self):
        self._algo_factories = []
        
    def add_algorithm_factory(self, algo_factory):
        self._algo_factories.append(algo_factory)
        
    def new(self, manifest, filename, st, callback):
        fcc = FileChecksumComputer(manifest, filename, st, callback)
        for af in self._algo_factories:
            fcc.add_algorithm(af.name, af.new())
        return fcc


class ProgressReporter:

    """Report progress to the user and estimate remaining time"""
    
    _max_measurement_age = 5.0 # in seconds
    _min_measurement_distance = 0.1 # in seconds
    
    def __init__(self, task):
        self._task = task
        self._measurements = []

    def remember_measurement(self, now, done):
        if not self._measurements or \
           now - self._measurements[-1][0] > self._min_measurement_distance:
            self._measurements.append((now, done))

    def forget_old_measurements(self, now):
        while self._measurements and \
              now - self._measurements[0][0] > self._max_measurement_age:
            del self._measurements[0]
        
    def estimate_speed(self):
        if len(self._measurements) >= 2:
            age_diff = self._measurements[-1][0] - self._measurements[0][0]
            assert age_diff > 0
            done_diff = self._measurements[-1][1] - self._measurements[0][1]
            assert done_diff >= 0, "task: " + repr(self._task)
            return done_diff / age_diff
        else:
            return 0
    
    def get_status(self):
        """Return current status: (total work, done so far, time estimate)"""
        now = time.time()
        (total, done) = self._task.estimate()
        self.remember_measurement(now, done)
        self.forget_old_measurements(now)
        speed = self.estimate_speed()
        assert speed >= 0
        if speed > 0:
            remaining_time = (total - done) / speed
        else:
            remaining_time = 0
        return total, done, remaining_time, speed


class ChecksumComputerQueue(BackgroundTask):

    def __init__(self, ui, base, manifest_type, algorithms, 
                 filenames_and_stats):
        self._ui = ui
    
        self._factory = FileChecksumComputerFactory()
        for algo in algorithms:
            algo_factory = ChecksumAlgorithmFactory(algo)
            self._factory.add_algorithm_factory(algo_factory)
            
        self._manifest = ManifestFactory(manifest_type).get_manifest(base)

        self._file_list = filenames_and_stats
        self._total = sum([x[1].st_size for x in self._file_list
                            if x[1] is not None])
        self._done = 0
        self._computer = None

    def get_manifest(self):
        return self._manifest

    def estimate(self):
        if self._computer:
            (this_total, this_done) = self._computer.estimate()
        else:
            this_done = 0
        done = self._done + this_done
        if self._total == 0:
            return done, done
        else:
            return self._total, done

    def work(self):
        if not self._computer and self._file_list:
            (filename, st) = self._file_list[0]
            self._computer = \
                self._factory.new(self._manifest, filename, st, 
                                  self.computer_callback)
            self._computer.setup()
        if self._computer and not self._computer.work():
            self._computer.cleanup()
            self._done += self._computer._bytes_read
            self._computer = None
            del self._file_list[0]
        return self._computer or self._file_list

    def computer_callback(self, manifest_file):
        raise Exception("Not implemented")


class ManifestCreator(ChecksumComputerQueue):

    def computer_callback(self, manifest_file):
        self._manifest.append(manifest_file)


class ManifestChecker(ChecksumComputerQueue):

    def __init__(self, ui, original_manifest):
        for mf in original_manifest[:]:
            filename = original_manifest.get_on_disk_filename(mf)
            try:
                mf.st = os.stat(filename)
            except os.error, e:
                ui.report_oserror(filename, e)
                original_manifest.remove(mf)
            else:
                if not stat.S_ISREG(mf.st.st_mode):
                    ui.report_nonfile(filename)
                    original_manifest.remove(mf)
        filenames_and_stats = \
            [(original_manifest.get_in_manifest_filename(mf), mf.st) 
                for mf in original_manifest]
        self._algos = sets.Set()
        for mf in original_manifest:
            for algo in mf.keys():
                self._algos.add(algo)
        ChecksumComputerQueue.__init__(self, ui, original_manifest._base,
                                       original_manifest._type, 
                                       self._algos,
                                       filenames_and_stats)
        self._original_manifest = original_manifest
        
    def computer_callback(self, manifest_file):
        for i in range(len(self._original_manifest)):
            omf = self._original_manifest[i]
            if omf.filenamex == manifest_file.filenamex:
                for algo in self._algos:
                    if algo in omf and algo in manifest_file:
                        if omf[algo] != manifest_file[algo]:
                            self._ui.report_failure(omf.filenamex)
                        else:
                            self._ui.report_success(omf.filenamex)
                        del self._original_manifest[i]
                        return
                raise Exception("Missing checksum for %s" % 
                                manifest_file.filenamex)
        raise Exception("Can't find %s in original manifest" % 
                        manifest_file.filenamex)


_md5sum_line_pattern = \
    re.compile(r"^(?P<hexdigest>[0-9a-fA-F]{32})\s+(?P<filename>.*\S)\s*$")
_sha1sum_line_pattern = \
    re.compile(r"^(?P<hexdigest>[0-9a-fA-F]{40})\s+(?P<filename>.*\S)\s*$")


class ManifestTypeGuesser:

    """Guess the type of a manifest file."""

    def __init__(self):
        self._guessers = [self._guess_md5sum, 
                          self._guess_sha1sum, 
                          self._guess_dsc,
                          self._guess_debian_changes,
                          self._guess_debian_sources,
                          self._guess_signed]
        
    def guess(self, buffer):
        """Guess the type, return None for unknown (maybe need more data?)"""
        for guesser in self._guessers:
            guess = guesser(buffer)
            if guess:
                return guess
        return None

    def _guess_md5sum(self, buffer):
        for line in buffer.split("\n"):
            if line and not _md5sum_line_pattern.search(line):
                return None
        return "md5"
        
    def _guess_sha1sum(self, buffer):
        for line in buffer.split("\n"):
            if line and not _sha1sum_line_pattern.match(line):
                return None
        return "sha1"
        
    def _guess_dsc(self, buffer):
        if buffer.startswith("Format: 1.0\n") and \
           "\nSource:" in buffer and \
           "\nBinary:" in buffer and \
           "\nFiles:" in buffer:
            return "dsc"
        
    def _guess_debian_changes(self, buffer):
        if buffer.startswith("Format: 1.7\n") and \
           "\nBinary:" in buffer and \
           "\nFiles:" in buffer:
            return "changes"

    def _guess_debian_sources(self, buffer):
        if buffer.startswith("Package:") and \
           "\nBinary:" in buffer and \
           "\nFormat: 1.0\n" in buffer and \
           "\nFiles:" in buffer:
            return "sources"

    def _guess_signed(self, buffer):
        if buffer.startswith("-----BEGIN PGP SIGNED MESSAGE-----\n"):
            return "signed"


class SimpleManifestParser:

    """Parse a simple manifest file of the md5sum or sha1sum type."""
    
    _pat = re.compile(r"^(?P<hexdigest>([0-9a-fA-F][0-9a-fA-F])+)\s+" +
                      r"(?P<filename>.*\S)\s*$", re.MULTILINE)

    def __init__(self, manifest, algorithm):
        self._buffer = ""
        self._manifest = manifest
        self._algo = algorithm

    def feed(self, data):
        """Feed some data to the parser"""
        self._buffer += data
        m = self._pat.search(self._buffer)
        while m:
            mf = ManifestFile(m.group("filename"), None)
            hexdigest = m.group("hexdigest")
            mf[self._algo] = hexdigest
            self._manifest.append(mf)
            self._buffer = self._buffer[m.end():]
            m = self._pat.search(self._buffer)
        
    def close(self):
        pass
        
    def get_manifest(self):
        """Return a Manifest with all files parsed out of the input"""
        return self._manifest


class Md5sumParser(SimpleManifestParser):

    def __init__(self, base):
        SimpleManifestParser.__init__(self, Md5Manifest(base), "md5")


class Sha1sumParser(SimpleManifestParser):

    def __init__(self, base):
        SimpleManifestParser.__init__(self, Sha1Manifest(base), "sha1")


class StringCursor:

    def __init__(self, string):
        self._string = string
        self._pos = 0
    
    def append(self, more_string):
        self._string += more_string

    def find(self, pattern, start=0):
        ret = self._string.find(pattern, self._pos + start)
        if ret == -1:
            return ret
        else:
            return ret - self._pos
        
    def startswith(self, pattern):
        return self._string.startswith(pattern, self._pos)
        
    def move(self, nchars):
        self._pos += nchars
        
    def __getitem__(self, index):
        return self._string[self._pos + index]
        
    def __getslice__(self, i, j):
        return self._string[self._pos + i:self._pos + j]

    def __len__(self):
        return len(self._string) - self._pos


class Rfc822LikeParser:

    """Base class for parsing files with a structure like RFC822 headers."""

    def __init__(self, manifest):
        self._buffer = StringCursor("")
        self._manifest = manifest
        
    def extract_full_header(self, final):
        if self._buffer.startswith("\n"):
            self.start_new_header_record()
            while self._buffer.startswith("\n"):
                self._buffer.move(1)
        physical_lines = []
        pos = 0
        while True:
            next_nl = self._buffer.find("\n", pos)
            if next_nl == -1:
                # Can't find the end of the line, so can't extract header.
                return None
            if not final and next_nl + 1 >= len(self._buffer):
                # The current line ends exactly where the buffer ends, so
                # we can't check if the next line continues the current
                # header.
                return None
            physical_lines.append(self._buffer[pos:next_nl + 1])
            pos = next_nl + 1
            if pos >= len(self._buffer) or not self._buffer[pos] in " \t":
                # The next line does not continue this header.
                break
        self._buffer.move(pos)
        return physical_lines

    def extract_all_full_headers(self, final):
        physical_lines = self.extract_full_header(final)
        while physical_lines:
            self.parse_full_header(physical_lines)
            physical_lines = self.extract_full_header(final)

    def start_new_header_record(self):
        pass

    def parse_full_header(self, physical_lines):
        pass

    def feed(self, data):
        self._buffer.append(data)
        self.extract_all_full_headers(False)

    def close(self):
        self.extract_all_full_headers(True)

    def get_manifest(self):
        return self._manifest


class DebianFilesParser(Rfc822LikeParser):

    """Base class for parsing Debian files (.dsc, .changes)."""

    def __init__(self, base):
        Rfc822LikeParser.__init__(self, Md5Manifest(base))

    def parse_full_header(self, physical_lines):
        if ":" not in physical_lines[0]:
            raise Exception("Header is missing name:\n" + 
                            repr("".join(physical_lines)))
        (name, rest) = physical_lines[0].split(":", 1)
        if name.lower() == "files":
            if rest.strip() != "":
                raise Exception("'Files:' header has " +
                                "garbage on header line:" + repr(rest))
            for line in physical_lines[1:]:
                self.parse_line_in_files_section(line)
        else:
            self.parse_other_header(name, [rest] + physical_lines[1:])

    def parse_other_header(self, name, contents):
        pass


class DscParser(DebianFilesParser):

    """Parse a Debian .dsc file."""
    
    def parse_line_in_files_section(self, line):
        parts = line.lstrip().split(" ", 2)
        if len(parts) != 3:
            raise Exception("'Files' has header malformed line: " + line)
        mf = ManifestFile(parts[2].strip(), None)
        mf["md5"] = parts[0]
        self._manifest.append(mf)


class DebianChangesFileParser(DebianFilesParser):

    """Parse a Debian .changes file."""
    
    def parse_line_in_files_section(self, line):
        parts = line.lstrip().split(" ", 4)
        if len(parts) != 5:
            raise Exception("'Files' has header malformed line: " + line)
        mf = ManifestFile(parts[4].strip(), None)
        mf["md5"] = parts[0]
        self._manifest.append(mf)


class DebianSourcesFileParser(DebianFilesParser):

    """Parse a Debian .changes file."""
    
    def __init__(self, base):
        DebianFilesParser.__init__(self, base)
        self._directory = None
    
    def parse_line_in_files_section(self, line):
        parts = line.lstrip().split(" ", 2)
        if len(parts) != 3:
            raise Exception("'Files' has header malformed line: " + line)
        mf = ManifestFile(os.path.join(self._directory, parts[2].strip()), 
                          None)
        mf["md5"] = parts[0]
        self._manifest.append(mf)

    def start_new_header_record(self):
        self._directory = None

    def parse_other_header(self, name, contents):
        if name == "Directory":
            self._directory = "".join(contents).strip()


class SignedFileParser:

    """Parse a file signed with PGP or GnuPG. Then re-parse the contents."""
    
    _pat = re.compile(r"-----BEGIN PGP SIGNED MESSAGE-----\n" +
                      r"(.+\n)*\n" +
                      r"(?P<contents>(.+\n)+)\n" +
                      r"-----BEGIN PGP SIGNATURE-----\n" +
                      r"(.*\n)*")
    
    def __init__(self, base):
        self._buffer = ""
        self._base = base
        
    def feed(self, data):
        self._buffer += data
        
    def close(self):
        m = self._pat.match(self._buffer)
        if not m:
            raise Exception("We can't seem to parse the signed file.")
        data = m.group("contents")
        guesser = ManifestTypeGuesser()
        contents_type = guesser.guess(data)
        if contents_type is None:
            raise Exception("Signed data has unknown type.")
        parser = ManifestParserFactory().get_parser(self._base, contents_type)
        parser.feed(data)
        parser.close()
        self._manifest = parser.get_manifest()
    
    def get_manifest(self):
        return self._manifest


class ManifestParserFactory:

    """Create new parsers for manifests."""
    
    _parsers = {
        "md5": Md5sumParser,
        "sha1": Sha1sumParser,
        "dsc": DscParser,
        "changes": DebianChangesFileParser,
        "sources": DebianSourcesFileParser,
        "signed": SignedFileParser,
    }
    
    def get_parser(self, base, manifest_type):
        if manifest_type in self._parsers:
            return self._parsers[manifest_type](base)
        else:
            raise Exception("Unknown manifest type for parsing: " + 
                            manifest_type)
